package net.knarcraft.ffmpegconverter.config;

import net.knarcraft.ffmpegconverter.property.MinimalSubtitlePreference;
import net.knarcraft.ffmpegconverter.utility.ConfigHelper;
import org.jetbrains.annotations.NotNull;

import java.io.IOException;
import java.util.List;

/**
 * A configuration used for FFMpegConvert
 */
public class Configuration {

    private final ConfigHandler configHandler;
    private boolean debug;
    private boolean useHardwareEncoding;
    private boolean useHardwareDecoding;
    private List<String> hardwareEncoders;
    private boolean alwaysEncodeFlac;
    private List<String> animeAudioLanguages;
    private List<String> animeSubtitleLanguages;
    private MinimalSubtitlePreference minimalSubtitlePreference;
    private boolean deInterlaceVideo;
    private boolean copyAttachedImages;
    private boolean overwriteFiles;

    /**
     * Instantiates and loads a new configuration
     */
    public Configuration() {
        this.configHandler = new ConfigHandler();
        try {
            this.configHandler.load();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }

        load();
    }

    /**
     * Loads/reloads this configuration
     */
    public void load() {
        try {
            this.configHandler.load();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }

        debug = ConfigHelper.asBoolean(configHandler.getValue(ConfigKey.DEBUG));
        useHardwareEncoding = ConfigHelper.asBoolean(configHandler.getValue(ConfigKey.USE_HARDWARE_ENCODING));
        useHardwareDecoding = ConfigHelper.asBoolean(configHandler.getValue(ConfigKey.USE_HARDWARE_DECODING));
        hardwareEncoders = ConfigHelper.asStringList(configHandler.getValue(ConfigKey.HARDWARE_ACCELERATED_ENCODERS));
        alwaysEncodeFlac = ConfigHelper.asBoolean(configHandler.getValue(ConfigKey.ENCODE_FLAC_ALWAYS));
        animeAudioLanguages = ConfigHelper.asStringList(configHandler.getValue(ConfigKey.AUDIO_LANGUAGES_ANIME));
        animeSubtitleLanguages = ConfigHelper.asStringList(configHandler.getValue(ConfigKey.SUBTITLE_LANGUAGES_ANIME));
        deInterlaceVideo = ConfigHelper.asBoolean(configHandler.getValue(ConfigKey.DE_INTERLACE_VIDEO));
        copyAttachedImages = ConfigHelper.asBoolean(configHandler.getValue(ConfigKey.COPY_ATTACHED_IMAGES));
        overwriteFiles = ConfigHelper.asBoolean(configHandler.getValue(ConfigKey.OVERWRITE_FILES));
        try {
            minimalSubtitlePreference = MinimalSubtitlePreference.valueOf(String.valueOf(configHandler.getValue(
                    ConfigKey.MINIMAL_SUBTITLE_PREFERENCE)));
        } catch (IllegalArgumentException | NullPointerException exception) {
            minimalSubtitlePreference = MinimalSubtitlePreference.AVOID;
        }
    }

    /**
     * Gets the underlying config handler for this configuration
     *
     * @return <p>The underlying config handler</p>
     */
    @NotNull
    public ConfigHandler getConfigHandler() {
        return this.configHandler;
    }

    /**
     * Gets whether debug mode is enabled
     *
     * @return <p>True if debug mode is enabled</p>
     */
    public boolean isDebugEnabled() {
        return this.debug;
    }

    /**
     * Gets whether hardware accelerated encoding is enabled
     *
     * @return <p>True if hardware accelerated encoding is enabled</p>
     */
    public boolean useHardwareEncoding() {
        return this.useHardwareEncoding;
    }

    /**
     * Gets whether hardware accelerated decoding is enabled
     *
     * @return <p>True if hardware accelerated decoding is enabled</p>
     */
    public boolean useHardwareDecoding() {
        return this.useHardwareDecoding;
    }

    /**
     * Gets whether FLAC audio tracks should always be re-encoded
     *
     * @return <p>Whether FLAC tracks should always be re-encoded</p>
     */
    public boolean alwaysEncodeFlac() {
        return this.alwaysEncodeFlac;
    }

    /**
     * Gets all hardware encoders defined in the configuration
     *
     * @return <p>The specified hardware encoders</p>
     */
    @NotNull
    public List<String> hardwareEncoders() {
        return this.hardwareEncoders;
    }

    /**
     * Gets the audio language priorities for usage when converting anime
     *
     * @return <p>The anime audio language priorities</p>
     */
    @NotNull
    public List<String> getAnimeAudioLanguages() {
        return this.animeAudioLanguages;
    }

    /**
     * Gets the subtitle language priorities for usage when converting anime
     *
     * @return <p>The anime subtitle language priorities</p>
     */
    @NotNull
    public List<String> getAnimeSubtitleLanguages() {
        return this.animeSubtitleLanguages;
    }

    /**
     * Gets the preference for minimal subtitles
     *
     * @return <p>The minimal subtitle preference</p>
     */
    @NotNull
    public MinimalSubtitlePreference getMinimalSubtitlePreference() {
        return this.minimalSubtitlePreference;
    }

    /**
     * Gets whether video streams should be de-interlaced
     *
     * @return <p>Whether to de-interlace streams</p>
     */
    public boolean deInterlaceVideo() {
        return this.deInterlaceVideo;
    }

    /**
     * Gets whether attached images should be copied
     *
     * <p>FFMpeg sometimes throws errors when including attached images.</p>
     *
     * @return <p>True if attached images should be copied</p>
     */
    public boolean copyAttachedImages() {
        return this.copyAttachedImages;
    }

    /**
     * Gets whether the original files should be overwritten after conversion has been finished
     *
     * @return <p>True if the original file should be overwritten</p>
     */
    public boolean overwriteFiles() {
        return this.overwriteFiles;
    }

}
