package net.knarcraft.ffmpegconverter.converter.sorter;

import net.knarcraft.ffmpegconverter.property.MinimalSubtitlePreference;
import net.knarcraft.ffmpegconverter.streams.AudioStream;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.List;

/**
 * A sorter for sorting/filtering subtitles by a minimal subtitle preference
 */
public class SpecialAudioSorter extends AbstractSorter<AudioStream> {

    private final MinimalSubtitlePreference minimalSubtitlePreference;

    /**
     * Instantiates a new special audio preference sorter
     *
     * @param minimalSubtitlePreference <p>The minimal subtitle preference sort/filter by</p>
     */
    public SpecialAudioSorter(@NotNull MinimalSubtitlePreference minimalSubtitlePreference) {
        this.minimalSubtitlePreference = minimalSubtitlePreference;
    }

    @Override
    public @NotNull List<AudioStream> sort(@NotNull List<AudioStream> input) {
        // Split all subtitles into full and minimal
        List<AudioStream> normalAudio = new ArrayList<>();
        List<AudioStream> specialAudio = new ArrayList<>();
        for (AudioStream audioStream : input) {
            if (audioStream.isSpecialAudio()) {
                specialAudio.add(audioStream);
            } else {
                normalAudio.add(audioStream);
            }
        }

        // Sort/filter subtitles based on full and minimal
        switch (this.minimalSubtitlePreference) {
            case REJECT -> {
                // Only return full subtitles
                return normalAudio;
            }
            case REQUIRE -> {
                // Only return minimal subtitles
                return specialAudio;
            }
            case NO_PREFERENCE -> {
                // Don't change order
                return input;
            }
            case PREFER -> {
                // Sort minimal subtitles first, and full subtitles last
                specialAudio.addAll(normalAudio);
                return specialAudio;
            }
            case AVOID -> {
                // Sort full subtitles first, and minimal subtitles last
                normalAudio.addAll(specialAudio);
                return normalAudio;
            }
            default -> throw new IllegalStateException("Unknown enum value encountered");
        }
    }

}
