package net.knarcraft.ffmpegconverter.converter.sorter;

import net.knarcraft.ffmpegconverter.streams.StreamObject;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.List;

/**
 * An interface describing a chaining-capable stream sorter
 *
 * @param <K> <p>The type of stream this sorter sorts</p>
 */
@SuppressWarnings("unused")
public interface StreamSorter<K extends StreamObject> {

    /**
     * Prepends this stream sorter to another
     *
     * <p>This stream sorter's next in chain will be set to other</p>
     *
     * @param other <p>The stream sorter to prepend</p>
     * @return <p>A reference to the first stream sorter in the current chain</p>
     */
    @NotNull
    StreamSorter<K> prepend(@NotNull StreamSorter<K> other);

    /**
     * Appends this stream sorter to another
     *
     * <p>This stream sorter is added to the end of other's chain</p>
     *
     * @param other <p>The stream sorter to append to this one</p>
     * @return <p>A reference to the first stream sorter in the current chain</p>
     */
    @NotNull
    StreamSorter<K> append(@NotNull StreamSorter<K> other);

    /**
     * Sorts the given input streams using this sorter only
     *
     * @param input <p>The input to sort</p>
     * @return <p>The sorted input</p>
     */
    @NotNull
    List<K> sort(@NotNull List<K> input);

    /**
     * Sorts the given input streams using all sorters in the chain
     *
     * @param input <p>The input to sort</p>
     * @return <p>The sorted input</p>
     */
    @NotNull
    List<K> chainSort(@NotNull List<K> input);

    /**
     * Gets whether this stream sorter has a sorter set as the next in its chain
     *
     * @return <p>True if a next chain item exists</p>
     */
    boolean hasChainElement();

    /**
     * Gets the next item in this stream sorter's chain
     *
     * @return <p>The next item in the chain</p>
     */
    @Nullable
    StreamSorter<K> getNextInChain();

    /**
     * Sets the given stream sorter as the next in the sorter chain
     *
     * @param next <p>The next item in the sorter chain</p>
     */
    void setNextInChain(@NotNull StreamSorter<K> next);

}
